%%% 3D split-step Fourie beam propagation method (BPM) for simulating
%%% tomographic phase microscope (TPM) by Choi et al.
%%% 1. Simulation of the TPM measurements with BPM
%%% 2. 3D reconstruction using iradon, non-negative constrain, ISTA-TV, and
%%%     FISTA-TV algorithms.
%%% U. S. Kamilov, BIG, EPFL, 2014.

%% Initialize
%%% Set up the parameters and define the computational grid.

clear; close all; clc;

lambda0 = 561e-9; % free space wavelength (m)
magnification = 100*200/180; % Mnominal*Fused/Fnominal
camera_pixel_size = 8e-6; % camera pixel size (m)

n0 = 1.518; % background refractive index (immersion oil)
k0 = 2*pi/lambda0; % free space wavenumber (1/m)
k = n0*k0; % medium wavenumber (1/m)

%%% depends on cropping
Nx = 256; % number of pixels along x
Ny = 256; % number of pixels along y
Nz = 128; % number of pixels along z

dx = camera_pixel_size/magnification; % discretization step along x
dy = camera_pixel_size/magnification; % discretization step along y
dz = camera_pixel_size/magnification; % discretization step along z

Lx = Nx*dx; % length of computational window along x
Ly = Ny*dy; % length of computational window along y
Lz = Nz*dz; % length of computational window along z

x = dx*(-Nx/2+1:Nx/2)'; % computational grid along x (horiz)
y = dy*(-Ny/2+1:Ny/2)'; % computational grid along y (vert)
z = dz*(1:Nz)'; % computational grid along z

[X, Y, Z] = meshgrid(x, y, z); % 3D meshgrid
[XX, YY] = meshgrid(x, y); %2D meshgrid

dkx = 2*pi/Lx; % frequency discretization step along x
dky = 2*pi/Ly; % frequency discretization step along y
dkz = 2*pi/Lz; % frequency discretization step along z

kx = dkx*[0:Nx/2-1, -Nx/2:-1]'; % frequency grid along x
ky = dky*[0:Ny/2-1, -Ny/2:-1]'; % frequency grid along y
kz = dkz*[0:Nz/2-1, -Nz/2:-1]'; % frequency grid along z

%[Kx, Ky, Kz] = meshgrid(kx, ky, kz); % 3D frequency meshgrid
[Kxx, Kyy] = meshgrid(kx, ky); % 2D frequency meshgrid

K2 = Kxx.^2+Kyy.^2; % frequency norm for all points

dphi = real(K2./(k+sqrt(k^2-K2))); % diffraction phase factor

[~, mid_index_y] = min(abs(y)); % midpoint index along x
[~, mid_index_x] = min(abs(x)); % midpoint index along y
[~, mid_index_z] = min(abs(z-Lz/2)); % midpoint index along z

%%% Create object
forwardObj = PlaneWave3D(Lx, Ly, Lz, Nx, Ny, Nz);

%% Phantom
%%% Generate a phantom consisting of a bead centered in the middle of the
%%% computation window.

dn = 0.03; % refractive index variation
n = n0+dn; % refractive index of the bead
D = 2e-6; % diameter of the bead
xc = 0; % center of the bead along x
yc = 0; % center of the bead along y
zc = Lz/2; % center of the bead along z

[~, ind_xc] = min(abs(x-xc)); % index of the bead center along x
[~, ind_yc] = min(abs(y-yc)); % index of the bead center along y
[~, ind_zc] = min(abs(z-zc)); % index of the bead center along z

R = sqrt((X-xc).^2 + (Y-yc).^2 + (Z-zc).^2);
f = dn*double(R < D/2); % final 3D phantom object
%sl = phantom3d(Nx/2);
%f = zeros(Ny, Nx, Nz);
%f(Ny/2-Ny/4+1:Ny/2+Ny/4, Nx/2-Nx/4+1:Nx/2+Nx/4, Nz/2-Nz/4+1:Nz/2+Nz/4)...
%    = sl;
f = dn*f/max(f(:));

%%% Plot the projection of the phantom along each of the spatial
%%% coordinates
figure(101);
set(101, 'Name', 'Object');
subplot(1, 3, 1);
imagesc(x, y, squeeze(mean(f, 3)));
hold on;
plot([xc xc], [y(1) y(end)], 'r--',...
    [x(1) x(end)], [yc yc], 'r--');
plot([x(mid_index_x) x(mid_index_x)], [y(1) y(end)], 'g:',...
    [x(1) x(end)], [y(mid_index_y) y(mid_index_y)], 'g:');
hold off;
axis equal tight;
xlabel('x');
ylabel('y');
title('Pz');
subplot(1, 3, 2);
imagesc(z, y, squeeze(mean(f, 2)));
hold on;
plot([zc zc], [y(1) y(end)], 'r--',...
    [z(1) z(end)], [yc yc], 'r--');
plot([z(mid_index_z) z(mid_index_z)], [y(1) y(end)], 'g:',...
    [z(1) z(end)], [y(mid_index_y) y(mid_index_y)], 'g:');
hold off;
axis equal tight;
xlabel('z');
ylabel('y');
title('Px');
subplot(1, 3, 3);
imagesc(z, x, squeeze(mean(f, 1)));
hold on;
plot([zc zc], [x(1) x(end)], 'r--',...
    [z(1) z(end)], [xc xc], 'r--');
plot([z(mid_index_z) z(mid_index_z)], [x(1) x(end)], 'g:',...
    [z(1) z(end)], [x(mid_index_x) x(mid_index_x)], 'g:');
hold off;
axis equal tight;
xlabel('z');
ylabel('x');
title('Py');


%% Measure
%%% Scan uniquely along the axis x and store the complex field at the end
%%% of the computational window

Ltheta = pi/6; % maximum scanning angle (radians)
Ntheta = 5; % number of scanning angles
thetas = linspace(-Ltheta/2, Ltheta/2, Ntheta);
dtheta = thetas(2)-thetas(1); % angle step

angs = thetas(end:-1:1)*180/pi-90; % for iradon in degrees

[~, mid_index_theta] = min(abs(thetas));

gob = zeros(Ny, Nx, Ntheta); % measurements when object is removed
gem = zeros(Ny, Nx, Ntheta); % measurements when object is present
gins = zeros(Ny, Nx, Ntheta); % measurements when object is present

for ind_theta = 1:Ntheta
    angx = thetas(ind_theta);
    angy = 0;
    
    Bx = 0.2*Lx; % beam scale (m)
    Bx = Bx/cos(angx); % projection of the propagation plane to XY plane
    
    By = 0.2*Ly;
    By = By/cos(angy);
    
    ain = exp(-((XX./Bx).^2 + (YY./By).^2)); % illumination beam amplitude
    pin = k*(sin(angx)*XX + sin(angy)*YY);
    
    gin = ain.*exp(1i*pin); % define a plane wave at the origin
    gin = ifft2(fft2(gin).*exp(-1i*(-Lz/2)*dphi)); % propagate to z = 0
        
    [uob, gtot_ob] = forwardObj.iforward(f, gin, angx);
    [uem, gtot_em] = forwardObj.iforward(zeros(size(f)), gin, angx);
    
    gins(:, :, ind_theta) = gin;
    gem(:,:,ind_theta) = uem; % store no object measurements
    gob(:,:,ind_theta) = uob; % store object measurements
    
    %%% Plot fields
    figure(102);
    set(102, 'Name', sprintf('[%d/%d] Propagation Profile',...
        ind_theta, Ntheta));
    
    subplot(1, 2, 1);
    imagesc(z, x, squeeze(abs(gtot_em(mid_index_y,:,:))));
    axis equal tight;
    xlabel('z');
    ylabel('x');
    title('Object Absent');
    
    subplot(1, 2, 2);
    imagesc(z, x, squeeze(abs(gtot_ob(mid_index_y,:,:))));
    hold on;
    plot([zc zc], [x(1) x(end)], 'r--',...
        [z(1) z(end)], [xc xc], 'r--');
    plot([z(mid_index_z) z(mid_index_z)], [x(1) x(end)], 'g:',...
        [z(1) z(end)], [x(mid_index_x) x(mid_index_x)], 'g:');
    viscircles([zc xc], D/2,...
        'DrawBackgroundCircle', false,...
        'EdgeColor', 'r',...
        'LineStyle', '--',...
        'LineWidth', 1);
    hold off;
    axis equal tight;
    xlabel('z');
    ylabel('x');
    title('Object Present');
    
    drawnow;
    
end

clear gin uem uob gtot_em gtot_ob;

%% Reconstruct with BPM

fhat = zeros(size(f));
stepSize = 2e-2;

numIter = 100;

costs = zeros(numIter, 1);
snrs = zeros(numIter, 1);

for iIter = 1:numIter
    
    grad = zeros(size(fhat));
    gout = zeros(size(gob));
    
    for ind_theta = 1:Ntheta
        %ind_theta = 1+mod(iIter-1, Ntheta);
    
        gin = gins(:, :, ind_theta);
        gi = gob(:, :, ind_theta);
    
        [igrad, igout] = forwardObj.icomputeGrad(gi, fhat, gin,...
            thetas(ind_theta));
        
        grad = grad + igrad;
        gout(:,:,ind_theta) = igout;
    end
    
    fhatnext = fhat - stepSize*grad/Ntheta;
    fhatnext(fhatnext<0) = 0;
    
    costs(iIter) = (1/Ntheta)*norm(gout(:)-gob(:))^2;
    snrs(iIter) = 20*log10(norm(f(:))/norm(f(:)-fhatnext(:)));
    diff = norm(fhatnext(:)-fhat(:))/norm(fhat(:));
    nmse = 20*log10(norm(fhatnext(:)-f(:))/norm(f(:)));
    
    fhat = fhatnext;
    
    fprintf('t = %d: cost = %e, nmse = %8.4f dB, diff. = %8.4f\n', iIter, costs(iIter), nmse, diff);
    
    % Plot the projection of the phantom along each of the spatial
    %%% coordinates
    figure(223);
    set(223, 'Name', sprintf('t = %d', iIter));
    subplot(2, 3, 1);
    imagesc(x, y, squeeze(mean(fhat, 3)));
    hold on;
    plot([xc xc], [y(1) y(end)], 'r--',...
        [x(1) x(end)], [yc yc], 'r--');
    plot([x(mid_index_x) x(mid_index_x)], [y(1) y(end)], 'g:',...
        [x(1) x(end)], [y(mid_index_y) y(mid_index_y)], 'g:');
    hold off;
    axis equal tight;
    xlabel('x');
    ylabel('y');
    title('Pz');
    subplot(2, 3, 2);
    imagesc(z, y, squeeze(mean(fhat, 2)));
    hold on;
    plot([zc zc], [y(1) y(end)], 'r--',...
        [z(1) z(end)], [yc yc], 'r--');
    plot([z(mid_index_z) z(mid_index_z)], [y(1) y(end)], 'g:',...
        [z(1) z(end)], [y(mid_index_y) y(mid_index_y)], 'g:');
    hold off;
    axis equal tight;
    xlabel('z');
    ylabel('y');
    title('Px');
    subplot(2, 3, 3);
    imagesc(z, x, squeeze(mean(fhat, 1)));
    hold on;
    plot([zc zc], [x(1) x(end)], 'r--',...
        [z(1) z(end)], [xc xc], 'r--');
    plot([z(mid_index_z) z(mid_index_z)], [x(1) x(end)], 'g:',...
        [z(1) z(end)], [x(mid_index_x) x(mid_index_x)], 'g:');
    hold off;
    axis equal tight;
    xlabel('z');
    ylabel('x');
    title('Py');
    
    subplot(2, 3, 4:5);
    semilogy(1:iIter, costs(1:iIter), 'b-',...
        iIter, costs(iIter), 'ro');
    xlim([1 numIter]);
    grid on;
    title(sprintf('Cost: %.2e', costs(iIter)));
    
    subplot(2, 3, 6);
    plot(1:iIter, snrs(1:iIter), 'b-',...
        iIter, snrs(iIter), 'ro');
    xlim([1 numIter]);
    grid on;
    title(sprintf('SNR: %.2f dB', snrs(iIter)));
    
    drawnow;
    
end