classdef PlaneWave3D < handle
    % Two-dimensional split-step Fourier beam propagation method for a
    % plane-wave beam.
    %
    % U. S. Kamilov, BIG, EPFL, 2014.
    
    properties(Constant)
        lambda0 = 561e-9; % Free-space wavelength [m]
        n0 = 1.518; % Background refractive index (immersion oil)
        k0 = 2*pi/PlaneWave3D.lambda0; % free space wavenumber [m-1]
        k = PlaneWave3D.n0*PlaneWave3D.k0; % medium wavenumber [m-1]
    end
    
    properties
        
        % computational grid (size of initial_focused_field)
        Nx;
        Ny;
        Nz;
        spectralFilterCenter;
        
        % discretization steps = (lambda0/NA)*(N_disk/N_detection_window)
        drx;
        dry;
        drz;
        
        % physical dimensions
        Lx;
        Ly;
        Lz;
        
        % domain coordinate vectors
        rx;
        ry;
        rz;
        
        % meshgrid of the vectors;
        %Rx;
        %Ry;
        %Rz;
        
        % 2d meshgrid
        Rxx;
        Ryy
        
        % discretization in the spatial frequency domain: dkx = 2*pi/Lx
        dkx;
        dky;
        dkz;
        
        % spatial frequency vectors
        kx;
        ky;
        kz;
        
        % phase factor for Fresnel diffraction
        dphi;
    end
    
    methods
        function this = PlaneWave3D(Lx, Ly, Lz, Nx, Ny, Nz)
            
            this.Lz = Lz;
            this.Lx = Lx;
            this.Ly = Ly;
            
            this.Nx = Nx;
            this.Ny = Ny;
            this.Nz = Nz;
            
            this.drx = this.Lx/this.Nx;
            this.dry = this.Ly/this.Ny;
            this.drz = this.Lz/this.Nz;
            
            this.rx = this.drx * (-this.Nx/2+1:this.Nx/2)';
            this.ry = this.dry * (-this.Ny/2+1:this.Ny/2)';
            this.rz = this.drz * (1:this.Nz);
            
            %[this.Rx, this.Ry, this.Rz] = meshgrid(this.rx,this.ry,this.rz);
            [this.Rxx, this.Ryy] = meshgrid(this.rx, this.ry);
            
            this.dkx = 2*pi/this.Lx;
            this.dky = 2*pi/this.Ly;
            this.dkz = 2*pi/this.Lz;
            
            this.kx = this.dkx * [0:this.Nx/2-1, -this.Nx/2:-1]';
            this.ky = this.dky * [0:this.Ny/2-1, -this.Ny/2:-1]';
            this.kz = this.dkz * [0:this.Nz/2-1, -this.Nz/2:-1]';
            
            [Kxx, Kyy] = meshgrid(this.kx, this.ky);
            
            K2 = Kxx.^2+Kyy.^2; % 2D frequency meshgrid
            
            this.dphi = real(K2./(this.k+sqrt(this.k^2-K2))); % diffraction phase factor
            
            [ixx , iyy] = meshgrid(1:this.Nx , 1:this.Ny);
            filterRadius = 70/2;
            
            this.spectralFilterCenter = zeros(this.Ny,this.Nx);
            this.spectralFilterCenter(sqrt((ixx-this.Nx/2-1).^2+(iyy-this.Ny/2-1).^2) < filterRadius) = 1;
            
        end
        
        function [gout, gtot] = iforward(this, x, gin, angle_inc)
            
            gtot = zeros(this.Nx, this.Ny, this.Nz);
            
            u = gin;
            
            for ind_z = 1:this.Nz
                u = ifft2(fft2(u).*exp(-1i*this.drz*this.dphi)); % diffraction step
                u = u.*exp(1i*this.k0*this.drz/cos(angle_inc)*x(:,:,ind_z)); % refraction step
                gtot(:,:,ind_z) = u;
            end
            
            %%% % Backpropagate to the center
            u = ifft2(fft2(u).*exp(-1i*(-this.Lz/2)*this.dphi));
            
            %%% % Low pass simultation to fit experimental data (see: hologram reconstruction algorithm)
            u = ifft2(ifftshift(fftshift(fft2(u./exp(1i*angle(gin)))).*this.spectralFilterCenter)).*exp(1i*angle(gin));
            
            gout = u;
        end
        
        function [grad, gout, gtot] = icomputeGrad(this, gi, x, gin, angle_inc)
            
            phi = this.dphi;
            
            [gout, gtot] = this.iforward(x, gin, angle_inc);
            
            grad = zeros(this.Nx,this.Ny,this.Nz);
            
            res = gout - gi;
            
            res = ifft2(ifftshift(fftshift(fft2((res./exp(1i*angle(gin))))).*this.spectralFilterCenter)).*exp(1i*angle(gin));
            
            res = ifft2(fft2(res).*exp(-1i*(this.Lz/2)*this.dphi));
            
            for ind_z = this.Nz:-1:2
                
                s = ifft2(fft2(gtot(:,:,ind_z-1)).*exp(-1i*this.drz*phi));
                s = conj(s).*res.*conj(1i*this.k0*this.drz/cos(angle_inc)...
                    *exp(1i*this.k0*this.drz/cos(angle_inc)*x(:,:,ind_z)));
                
                grad(:,:,ind_z) = real(s);
                
                res = res.*conj(exp(1i*this.k0*this.drz/cos(angle_inc)*x(:,:,ind_z)));
                res = ifft2(fft2(res) .* exp(1i*this.drz*phi));
            end
            
            s = ifft2(fft2(gin).*exp(-1i*this.drz*phi));
            s = conj(s).*res...
                .*conj(1i*this.k0*this.drz/cos(angle_inc)*exp(1i*this.k0*this.drz/cos(angle_inc)*x(:,:,1)));
            grad(:,:,1) = real(s);
        end
        
        function gouthat = computeFullForward(this, fhat, gin, angles)
            gouthat = zeros(size(gin));
            Ntheta = size(gouthat, 3);
            
            for ind_theta = 1:Ntheta
                obj = this;
                iangle = angles(ind_theta);
                igin = gin(:,:,ind_theta);
                igout = obj.iforward(fhat, igin, iangle);
                gouthat(:,:,ind_theta) = igout;
                fprintf('computeFullForward: %d of %d\n',...
                    ind_theta, Ntheta);
            end
        end
        
        function [grad, gouthat] = computeFullGrad(this, gout, fhat, gin, angles)
            
            grad = zeros(size(fhat));
            gouthat = zeros(size(gout));
            Ntheta = size(gout,3);
            
            for ind_theta = 1:Ntheta
                obj = this;
                iangle = angles(ind_theta);
                igin = gin(:,:,ind_theta);
                igout = gout(:, :, ind_theta);
                [igrad, igouthat] = obj.icomputeGrad(igout, fhat, igin, iangle);
                grad = grad + igrad;
                gouthat(:,:,ind_theta) = igouthat;
                fprintf('computeFullGradient: %d of %d\n',...
                    ind_theta, Ntheta);
            end
            grad = grad/Ntheta;
        end
    end
end